#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-11-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000  

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions
  
def gegl_softglow (image, drawable, uradius, ubright, usharp):    
    glow_radius=float(uradius)  #10.0
    brightness=float(ubright)   # 0.3
    sharpness=float(usharp)     # 0.85
    gegl_graph_string="gegl:softglow glow-radius=%f brightness=%f sharpness=%f" % (glow_radius, brightness, sharpness)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_median_blur (image, drawable, uradius):
    radius=float(uradius)
    percentile=50.0
    alpha_percentile=50.0
    gegl_graph_string="gegl:median-blur radius=%f percentile=%f alpha-percentile=%f" % (radius, percentile, alpha_percentile)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_gaussian_blur (image, drawable, ux, uy):    # gaussian blur  
    std_dev_x=float(ux)    #1.50     #user_range   (0.24, 100.0)
    std_dev_y=float(uy)    #1.50     #user_range   (0.24, 100.0)
    mystring="gegl:gaussian-blur std-dev-x=%f std-dev-y=%f"
    gegl_graph_string=mystring % (std_dev_x, std_dev_y)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_hue_chroma (image, drawable, uhue, uchroma, ulightness):    # Hue-Chroma  
    hue=uhue                #0.0   #value_range  (-180.0, 180.0) 
    chroma=uchroma          #12.0  #value_range  (-100.0, 100.0)  
    lightness=ulightness    #0.0   #value_range  (-100.0, 100.0)
    mystring="gegl:hue-chroma hue=%d chroma=%d lightness=%d"
    gegl_graph_string=mystring % (hue, chroma, lightness)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_saturation (image, drawable, uscale, ucolorspace):    # saturation 
    scale=uscale            #1.5  #user_range   (0.0, 2.0)
    colorspace=ucolorspace  #0    #value_range  (0, 1, 2)  Native, CIE-lab/lch, CIE-iuv
    mystring="gegl:saturation scale=%f colorspace=%s"
    gegl_graph_string=mystring % (scale, colorspace)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_exposure (image, drawable, ublack, uexpo):      
    black_level=ublack    # value_range (-0.1, 0.1)
    exposure=uexpo        # ui_range    (-10.0, 10.0)
    gegl_graph_string="gegl:exposure black-level=%f exposure=%f" % (black_level, exposure)
    gegl_graph(image, drawable,gegl_graph_string)
    
def gegl_unsharp_mask (image, drawable, usdev , uscale, uthreshold):      
    std_dev=usdev           #user_range   (0.0, 40.0)
    scale=uscale            #user_range   (0.0, 10.0)
    threshold=uthreshold    #user_range   (0.0, 1.0) 
    mystring="gegl:unsharp-mask std-dev=%f scale=%f threshold=%f"
    gegl_graph_string=mystring % (std_dev, scale, threshold)
    gegl_graph(image, drawable,gegl_graph_string)

############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

#---------------------------------------------------------------------------
# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer, meanVadj, meanSadj, equaladj, rmvextremes):
    
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    
    # adjusting values
    newBG = adjustmeanSV (newImage, newBG, meanVadj, meanSadj)
    newBG = equalizeV (newImage, newBG, equaladj)
    
    # remove extremes
    pdb.gimp_drawable_curves_spline (newBG, HISTOGRAM_VALUE, 6,
                                     (0.00,float(rmvextremes*0.01),
                                      0.50,0.50,
                                      1.00,float(1.00-rmvextremes*0.01)))
    newBG.opacity = 100.0
    
    return(newImage)
    
#---------------------------------------------------------------------------
# adjust mean values and mean saturation
#---------------------------------------------------------------------------
def adjustmeanSV (inImage, inLayer, meanVadj, meanSadj):
    # decompose HSV
    (HSVI,HL,SL,VL) = pdb.plug_in_decompose (inImage, inLayer, "HSV", True)
    #pdb.gimp_display_new (HSVI)
    #info_msg ("meanVadj="+str(meanVadj)+", meanSadj="+str(meanSadj))
    HL=HSVI.layers[0]; SL=HSVI.layers[1]; VL=HSVI.layers[2];
    if meanSadj!=0:
        #info_msg ("calling adjustmeanv for saturation")
        adjustmeanv (HSVI, SL, meanSadj)
    if meanVadj!=0:
        #info_msg ("calling adjustmeanv for value")
        adjustmeanv (HSVI, VL, meanVadj)
    pdb.plug_in_recompose (HSVI, HL)
    return(inLayer)

#---------------------------------------------------------------------------
# adjust mean values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer, meanadj):
    # take care of the existing mean in determining the theoretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    thmean = (127.5*meanadj/100+meanl*(100.0-meanadj)/100)   
    #info_msg ("thmean="+str(thmean))
    # range for the loops
    minmean = thmean*1.05  
    maxmean = thmean*1.15  
    
    modified=False
    if meanl < minmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        loop=0
        while meanl < minmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        loop=0
        while meanl > maxmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))

    return(inLayer)
    
#---------------------------------------------------------------------------
# equalize values  
#---------------------------------------------------------------------------
def equalizeV (inImage, inLayer, equaladj):
    equalized = inLayer.copy()
    inImage.add_layer (equalized, 0)
    pdb.gimp_drawable_equalize (equalized, False)
    equalized.mode = LAYER_MODE_SOFTLIGHT
    equalized.opacity = equaladj
    inLayer = pdb.gimp_image_merge_down (inImage, equalized, CLIP_TO_BOTTOM_LAYER)
    return(inLayer)

#---------------------------------------------------------------------------
# adjust values merging inverted
#---------------------------------------------------------------------------
def mergeinvert (inImage, inLayer, meanadj):
    
    #check luminance
    inLayer = adjustmeanv (inImage, inLayer, meanadj)
    vinverted = inLayer.copy()
    inImage.add_layer (vinverted, 0)
    pdb.plug_in_vinvert (inImage, vinverted)
    vinverted.mode = LAYER_MODE_SOFTLIGHT
    inLayer = pdb.gimp_image_merge_down (inImage, vinverted, CLIP_TO_BOTTOM_LAYER)
    # remove extremes
    pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 6,
                                     (0.00,0.05,0.50,0.50, 1.00,0.95))
    # - USM
    uSdev=3.0; uScale=1.0; uThreshold=0.0       
    gegl_unsharp_mask (inImage, inLayer, uSdev, uScale, uThreshold)
    
    return(inLayer)
    
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def JvidEffectU (inImage, inLayer, widthheight,   
                 blurval, meanVadj, meanSadj, equaladj, rmvextremes,
                 doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
        
    OrigWH = inLayer.width+inLayer.height
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    newImage = createOutimage (inImage, inLayer,
                               meanVadj, meanSadj, equaladj, rmvextremes)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # prepare image for effects Blur + USM 
    # - Blur  
    Blurred = Preprocessed.copy()
    newImage.add_layer (Blurred, 0)
    uRadiusX=float(blurval); uRadiusY=float(blurval)
    Blurred.name = "Blur applied "+str(blurval) 
    gegl_gaussian_blur (newImage, Blurred, uRadiusX, uRadiusY)
    # - USM
    uSdev=float(blurval); uScale=1.0; uThreshold=0.0       # new parameters
    Blurred.name = "Gaussian Blur "+str(uSdev)+" then USM "+str(uSdev)  
    gegl_unsharp_mask (newImage, Blurred, uSdev, uScale, uThreshold)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # Effect: B&W Emboss gcd_emboss 159,0
    (Mean,B,C,D,E,F) = pdb.gimp_drawable_histogram (Blurred, HISTOGRAM_VALUE, 0.0,1.0)
    #info_msg ("mean for Emboss="+str(Mean))
    pdb.plug_in_gmic_qt(newImage, Blurred, 1,2,
    "-v - -gcd_emboss "+str(Mean)+",0 ")
    Embossed = newImage.layers[0]
    Embossed.mode = LAYER_MODE_HARDLIGHT
    Embossed.opacity = 50.0 #80.0  
    Embossed.name = "[GMIC] gcd_emboss "+str(Mean)+",0 "
    # remove extremes
    pdb.gimp_drawable_curves_spline (Embossed, HISTOGRAM_VALUE, 6,
                                     (0.00,float(rmvextremes*0.01),
                                      0.50,0.50,
                                      1.00,float(1.00-rmvextremes*0.01)))
    #pdb.gimp_display_new (newImage) 
    #return
    
    # Input for further step
    BWE = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (BWE, 0)
    BWE.name = "Embossed "
    BWE.mode = LAYER_MODE_OVERLAY   #PIN_LIGHT
    pdb.gimp_item_set_visible (Embossed, False)
    
    #pdb.gimp_display_new (newImage)    
    #return
    
    BWE2 = BWE.copy()
    newImage.add_layer (BWE2, 0)
    BWE2.name = "BWE copy "
    BWE2.mode = LAYER_MODE_MULTIPLY
    BWE2.opacity = Mean/2.56 #75.0
    (MeanE,B,C,D,E,F) = pdb.gimp_drawable_histogram (BWE, HISTOGRAM_VALUE, 0.0,1.0)
    if MeanE<85: BWE2.mode = LAYER_MODE_SCREEN
    elif MeanE>170: BWE2.mode = LAYER_MODE_MULTIPLY
    else: BWE2.mode = LAYER_MODE_NORMAL # no impact   
    
    # Input for Dream Smoothing
    Base2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base2, 0)
    Base2.name = "Base for Dreamsmooth "
    Base2.mode = LAYER_MODE_NORMAL
    Base2 = adjustmeanv (newImage, Base2, meanVadj)  
    Base2E = Base2.copy()   # try adding Equalize a bit
    newImage.add_layer (Base2E, 0)
    Base2E = equalizeV (newImage, Base2E, equaladj)
    #pdb.gimp_display_new (newImage)    
    #return
    Base2E.opacity = 50.0   #30.0
    Base2E.mode = LAYER_MODE_SOFTLIGHT  #NORMAL
    Base2 = pdb.gimp_image_merge_down (newImage, Base2E, CLIP_TO_IMAGE)
    
    #pdb.gimp_display_new (newImage)    
    #return
    
    # Effect Dreamsmooth  
    # [G'MIC]  fx_dreamsmooth 2,1,1,0.8,0,0.8,1,24,0
    pdb.plug_in_gmic_qt(newImage, Base2, 1,2 ,
    "-v - -fx_dreamsmooth 2,1,1,0.8,0,0.8,1,24,0 ")
    Dreams = newImage.layers[0]
    Dreams.mode = LAYER_MODE_MULTIPLY #tut SOFTLIGHT   
    Dreams.opacity = 50.0  
    Dreams.name = "[GMIC] -fx_dreamsmooth 2,1,1,0.8,0,0.8,1,24,0"
    
    #pdb.gimp_display_new (newImage) 
    #return
    ''' # moved at end
    # Input for Canvas 
    '''
    # Pre-outcome to check luminosity
    Pre_outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
    newImage.add_layer (Pre_outcome, 0)
    Pre_outcome.name = "Preliminary outcome to check luminosity"
    Pre_outcome.mode = LAYER_MODE_NORMAL  
    # adjust values
    Pre_outcome = mergeinvert (newImage, Pre_outcome, blurval)
    Pre_outcome.mode = LAYER_MODE_HARDLIGHT #MULTIPLY
    Pre_outcome.opacity=25.0
    # remove extremes
    pdb.gimp_drawable_curves_spline (Pre_outcome, HISTOGRAM_VALUE, 6,
                                     (0.00,float(rmvextremes*0.01),
                                      0.50,0.50,
                                      1.00,float(1.00-rmvextremes*0.01)))
    
    # Final Effect: B&W Emboss gcd_emboss  
    (Mean,B,C,D,E,F) = pdb.gimp_drawable_histogram (Pre_outcome, HISTOGRAM_VALUE, 0.0,1.0)
    #info_msg ("mean for Emboss="+str(Mean))
    pdb.plug_in_gmic_qt(newImage, Pre_outcome, 1,2,
    "-v - -gcd_emboss "+str(Mean)+",0 ")
    FinalEmboss = newImage.layers[0]
    FinalEmboss.mode = LAYER_MODE_SOFTLIGHT #LINEAR_LIGHT #GRAIN_MERGE
    FinalEmboss.opacity = 50.0  #25.0  
    FinalEmboss.name = "[GMIC] gcd_emboss "+str(Mean)+",0 "
    adjustmeanv (newImage, FinalEmboss, meanVadj)  # added
    uRadiusX=float(width*20.0/FixWH); uRadiusY=float(height*20.0/FixWH)
    gegl_gaussian_blur (newImage, FinalEmboss, uRadiusX, uRadiusY)
    
    # Input for Canvas ### moved here from above
    Base3 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base3, 0)
    Base3.name = "Base For Canvas "
    Base3.mode = LAYER_MODE_NORMAL
    
    pdb.plug_in_gmic_qt(newImage, Base3, 1,0,       # was 1,2
    "-v - -texturize_canvas 20,4,0.8 ")
    Canvas = newImage.layers[0]
    Canvas.mode = LAYER_MODE_SCREEN #NORMAL
    Canvas.opacity = 50.0  
    Canvas.name = "applied canvas 20,4,0.8  "
    # remove extremes
    pdb.gimp_drawable_curves_spline (Canvas, HISTOGRAM_VALUE, 6,
                                     (0.00,float(rmvextremes*0.01),
                                      0.50,0.50,
                                      1.00,float(1.00-rmvextremes*0.01)))
    
    #pdb.gimp_display_new (newImage)    
    #return
    #pdb.gimp_display_new (newImage) 
    #return
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        pdb.gimp_image_remove_layer (newImage, Embossed)
        outcome.name = "outcome"
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect U (Graphic_Effect_Fake_3d), parms= "
    outcome.name = outcome.name+", "+str(blurval)
    outcome.name = outcome.name+", "+str(meanVadj)
    outcome.name = outcome.name+", "+str(meanSadj)
    outcome.name = outcome.name+", "+str(equaladj)
    outcome.name = outcome.name+", "+str(rmvextremes)
    
    if widthheight==0: widthheight=OrigWH
    uRatio = widthheight*1.00/FixWH*1.00
    uWidth = int(width*uRatio)
    uHeight = int(height*uRatio)
    #info_msg ("SCALING")   # not during tests, scaling is slow
    pdb.gimp_image_scale (newImage, uWidth, uHeight)   # to user chosen size
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect U (Graphic_Effect_Fake_3d) "

register(
    "JvidEffectU",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_ADJUSTMENT, "softglow", "Blur amount", 1.7, (1.0, 5.0, 0.1)),
        (PF_ADJUSTMENT, "meanva", "Mean value adj", 50.0, (0.0, 100.0, 5.0)),
        (PF_ADJUSTMENT, "satuva", "Saturation value adj", 5.0, (-50.0, 100.0, 5.0)),
        (PF_ADJUSTMENT, "equala", "Equalize adj", 20.0, (0.0, 100.0, 5.0)),
        (PF_ADJUSTMENT, "noextreme", "Remove extreme values?"+"\n (7=good for people)",
                                        5.0, (0.0, 10.0, 0.1)),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectU,
    menu=menu
)

main()
